
import Prelude
import Parser
import TextUtil
import Maybe
import List
import RegExpData
import Subset
import System


main :: IO ()
main = do args <- getArgs
          rules <- readFile "rules.txt"
          putStr (main2 args rules)

exec :: String -> IO()
exec arg = do rules <- readFile "rules.txt"
              putStr (main2 [arg] rules)

execs :: String -> IO()
execs arg = do rules <- readFile "rules.txt"
               putStr (main2 ["-v", arg] rules)


main2 :: [String] -> String -> String
main2 args rule_source = answer
    where
        rules = map parseRule (lines (rule_source))
        verbose = any (== "-v") args
        regex = head (filter (\x -> head x /= '-') args)
        result = simplify rules (parse regex)
        answer = if verbose then unlines (map show result) else show (last result)



simplify :: [Rule] -> RegExp -> [RegExp]
simplify rules reg = reg : (if isJust res then simplify rules (represent (fromJust res)) else [])
    where res = applyOne rules (represent reg)


represent (Concat [x]) = x
represent (Concat xs ) = Concat (hoist f xs)
    where f (Concat x) = x
          f x          = [x]

represent (Union  [x]) = x
represent (Union   xs) = Union (hoist f xs)
    where f (Union  x) = x
          f x          = [x]

represent x            = x


hoist :: (a -> [a]) -> [a] -> [a]
hoist f = concat . map f





applyOne :: [Rule] -> RegExp -> Maybe RegExp
applyOne rules reg = zipJust (repeat id) (map (`simplif` reg) rules)


simplif :: Rule -> RegExp -> Maybe RegExp
simplif x y@(Terminal a) = apply x y
simplif x y@(Omega     ) = apply x y

simplif x y@(Kleene a) = zipJust [id, Kleene] [apply x y, simplif x a]

simplif x y@(Concat [a]) = simplif x a
simplif x y@(Union  [a]) = simplif x a

simplif x y@(Concat as) = zipJust [id, (\x -> Concat [head as, x]), (\x -> Concat [x, last as])]
                                  [apply x y, simplif x (Concat (tail as)), simplif x (Concat (init as))]

simplif x y@(Union  as) = zipJust (id: map (\x y -> Union [x, y]) as)
                                  (apply x y: map (simplif x . Union) (subsets as))




--all subsets where length is one less
subsets :: [a] -> [[a]]
subsets [] = []
subsets (a:as) = as : map (a:) (subsets as)





zipJust :: [(a -> a)] -> [Maybe a] -> Maybe a
zipJust _ [] = Nothing
zipJust (f:_) (Just x:_) = Just (f x)
zipJust (_:fs) (Nothing:xs) = zipJust fs xs


apply :: Rule -> RegExp -> Maybe RegExp
apply (Rule find check rep) x = if null res then Nothing else Just (generate (head res) rep)
    where
        res = filter f (bind find x)
        f y = not (null y) && all (g (generate y)) check
        
        g q (Ewp a) = ewp (q a)
        g q (Equal a b) = q a == q b
        g q (Subset a b) = q a `issub` q b



generate :: [(Char, RegExp)] -> RegExp -> RegExp
generate table reg = f reg
    where
        f (Omega     ) = Omega
        f (Kleene   a) = Kleene (f a)
        f (Concat   a) = Concat (map f a)
        f (Union    a) = Union (map f a)
        f (Terminal a) = if isJust symbol then fromJust symbol else error ("Could not resolve '" ++ [a] ++ "' in: " ++ show reg ++ ", using: " ++ show table)
            where symbol = lookup a table



-- rule, data
bind :: RegExp -> RegExp -> [[(Char, RegExp)]]
bind (Terminal a) x = [[(a, x)]]
bind (Kleene   a) (Kleene b) = bind a b
bind (Concat   a) (Concat b) = bindList associative Concat a b
bind (Union    a) (Union  b) = bindList associative_commutative Union a b
bind (Omega     ) (Omega   ) = [[]]
bind _ _ = []


bindList :: (Int -> [RegExp] -> [[[RegExp]]]) -> ([RegExp] -> RegExp) -> [RegExp] -> [RegExp] -> [[(Char, RegExp)]]
bindList ac reg rule x = catMaybes (map f (nonempty (ac (length rule) x)))
    where
        f y = if any null res then Nothing else Just (concat (concat res))
            where res = zipWith g rule y
        
        g rule1 x1 = bind rule1 (if null (tail x1) then head x1 else reg x1)



nonempty :: [[[a]]] -> [[[a]]]
nonempty x = filter f x
    where
        f y = all (not . null) y


associative :: Int -> [a] -> [[[a]]]
associative 1 x = [[x]]
associative n x = concat (zipWith f (inits x) (map (associative (n-1)) (tails x)))
    where
        f :: [a] -> [[[a]]] -> [[[a]]]
        f hd tl = map (hd:) tl


commutative :: [a] -> [[a]]
commutative [] = [[]]
commutative [a] = [[a]]
commutative (x:xs) = concat (map g (commutative xs))
    where
        g x = zipWith f (inits x) (tails x)
        f pre post = pre ++ [x] ++ post


associative_commutative :: Int -> [a] -> [[[a]]]
associative_commutative n x = concat (map (associative n) (commutative x))



parseRule :: String -> Rule
parseRule x = Rule leftExp2 (conds2 ++ conds) rightExp
    where
        (leftExp2, conds2) = unify leftExp
    
        Just (left, right) = splitPair "=" x
        mayLeft = splitPair "|" left
        Just (leftReg, leftCond) = mayLeft
        leftExp = parse (trim (if isJust mayLeft then leftReg else left))
        conds = if isJust mayLeft then parseConds (trim leftCond) else []
        rightExp = parse (trim right)


unify :: RegExp -> (RegExp, [Condition])
unify x = (renExp newList x, condlist newList oldList)
    where
    
        oldList = terminals x
        newList = renlist [] oldList (['a'..'z'] \\ oldList)
    
        terminals (Terminal x) = [x]
        terminals (Omega     ) = []
        terminals (Kleene x  ) = terminals x
        terminals (Concat x  ) = concat (map terminals x)
        terminals (Union  x  ) = concat (map terminals x)
        
        
        renlist used [    ] avail = []
        renlist used (x:xs) avail =
            if x `elem` used
            then head avail : renlist used xs (tail avail)
            else x : renlist (x:used) xs avail
            
        
        condlist []     []     = []
        condlist (a:as) (b:bs) = (if a == b then [] else [Equal (Terminal a) (Terminal b)])
                                 ++ condlist as bs
        
        renExp [a] (Terminal x) = Terminal a
        renExp [ ] (Omega     ) = Omega
        renExp as  (Kleene   x) = Kleene (renExp as x)
        renExp as  (Concat   x) = Concat (renMap as x)
        renExp as  (Union    x) = Union  (renMap as x)
        
        
        renMap [] []     = []
        renMap as (x:xs) = renExp (take n as) x : renMap (drop n as) xs
            where n = length (terminals x)



data Rule = Rule RegExp [Condition] RegExp
            deriving (Show)

data Condition = Ewp RegExp
               | Subset RegExp RegExp
               | Equal RegExp RegExp
               deriving (Show)


parseConds :: String -> [Condition]
parseConds x = map (parseCond . trim) (splitList "," x)

parseCond :: String -> Condition
parseCond x =
        case name of
            "sub" -> Subset a2 b2
            "ewp" -> Ewp a1
            _     -> error ("Invalid Condition: " ++ name)
    where
        (name:params) = splitList " " x
        ps = map parse params
        [a1] = ps
        [a2, b2] = ps



nightmare = "f+^^*f+(a+^^*a)(^+0^*a)*(0+0^*f)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))" -- (^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f))+(f+^^*f+(a+^^*a)(^+0^*a)*(0+0^*f)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))" -- *(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))+(0+^^*0+(a+^^*a)(^+0^*a)*(b+0^*0)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))+(f+^^*f+(a+^^*a)(^+0^*a)*(0+0^*f)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(0+0^*0+(0+0^*a)(^+0^*a)*(b+0^*0)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))))(^+c^*0+(0+c^*a)(^+0^*a)*(b+0^*0)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))+(0+c^*f+(0+c^*a)(^+0^*a)*(0+0^*f)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(0+0^*0+(0+0^*a)(^+0^*a)*(b+0^*0)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))))*(0+c^*f+(0+c^*a)(^+0^*a)*(0+0^*f)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f))+(0+c^*f+(0+c^*a)(^+0^*a)*(0+0^*f)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f))))+0"


fnightmare = "f+^^*f+(a+^^*a)(^+0^*a)*(0+0^*f)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f))+(f+^^*f+(a+^^*a)(^+0^*a)*(0+0^*f)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))+(0+^^*0+(a+^^*a)(^+0^*a)*(b+0^*0)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))+(f+^^*f+(a+^^*a)(^+0^*a)*(0+0^*f)+(d+^^*d+(a+^^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(0+0^*0+(0+0^*a)(^+0^*a)*(b+0^*0)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))))(^+c^*0+(0+c^*a)(^+0^*a)*(b+0^*0)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))+(0+c^*f+(0+c^*a)(^+0^*a)*(0+0^*f)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(0+0^*0+(0+0^*a)(^+0^*a)*(b+0^*0)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*0+(0+e^*a)(^+0^*a)*(b+0^*0))))*(0+c^*f+(0+c^*a)(^+0^*a)*(0+0^*f)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f))+(0+c^*f+(0+c^*a)(^+0^*a)*(0+0^*f)+(0+c^*d+(0+c^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f)))*(^+0^*f+(0+0^*a)(^+0^*a)*(0+0^*f)+(0+0^*d+(0+0^*a)(^+0^*a)*(0+0^*d))(^+e^*d+(0+e^*a)(^+0^*a)*(0+0^*d))*(0+e^*f+(0+e^*a)(^+0^*a)*(0+0^*f))))+0"

death = "^*f+(^*d(^+e^*d)*e^*a^*b+^*a^*b)(^+c^*d(^+e^*d)*e^*a^*b+c^*a^*b)*(c^*f+c^*d(^+e^*d)*e^*f+(c^*f+c^*d(^+e^*d)*e^*f)^*)+(^*f+^*d(^+e^*d)*e^*f)^*+^*d(^+e^*d)*e^*f"
death2 = "^*f+^*d(^+e^*d)*e^*f+^*(f+d(^+e^*d)*e^*f)^*+^*(d(^+e^*d)*e^*a^*b+a^*b)(^+c^*(d(^+e^*d)*e^*a^*b+a^*b))*(c^*f+c^*d(^+e^*d)*e^*f+c^*(f+d(^+e^*d)*e^*f)^*)"
death_instant = "^*(d*e^*+^*+(d*e^*a^*b+a^*)*^*(d*e^*f^*))"

